#ifdef WIN32
#include <time.h>
#else
#include <sys/time.h>
#include <unistd.h>
#endif

#ifdef HAVE_GLX
#include "vroot.h"
#endif

#include "main.h"

#include <gfx/quat.h>

#define MIN_DELAY 5

GLCanvas::GLCanvas() : need_refresh(true), sdl_window(true), spf(1.0/30.0)
{
#ifdef HAVE_GLX
    window = 0;
    display = 0;
    window_id = 0;
#endif
    width = height = 0;
    full_screen = false;
}

GLuint GLCanvas::get_millis()
{
#ifdef HAVE_SDL
    return SDL_GetTicks();
#else
    struct timeval tv;
    gettimeofday(&tv, 0);
    return (1000*tv.tv_sec + tv.tv_usec/1000);
#endif
}

int GLCanvas::create_window()
{
#ifdef HAVE_GLX
    if (!sdl_window) {
	if ((display = XOpenDisplay(0)) == 0)
	    return -1;
	screen = DefaultScreen(display);

	window = window_id ? window_id : RootWindow(display, screen);
	XWindowAttributes wa;
	XGetWindowAttributes(display, window, &wa);
	Visual *visual = wa.visual;

	XVisualInfo templ;
	templ.screen = screen;
	templ.visualid = XVisualIDFromVisual(visual);
	int out_count;
	XVisualInfo *vinfo = XGetVisualInfo(display,
		VisualScreenMask | VisualIDMask, &templ, &out_count);
	if (!vinfo)
	    return -1;

	GLXContext context = glXCreateContext(display, vinfo, 0, GL_TRUE);
	XFree(vinfo); 
	if (!context)
	    return -1;

	if (!glXMakeCurrent(display, window, context))
	    return -1;

	XMapRaised(display, window);

	return 0;
    }
#endif
#ifdef HAVE_SDL
    if (sdl_window) {
	int ret;
	if ((ret=SDL_Init(SDL_INIT_VIDEO)) < 0)
	    return ret;

	SDL_Rect **modes;
	modes = SDL_ListModes(0, SDL_FULLSCREEN|SDL_OPENGL);
	if (modes == (SDL_Rect**)0) {
	    cerr << "ERROR: no modes available" << endl;
	    return -1;
	}
	if (modes == (SDL_Rect**)-1) {
	    width = 1024;
	    height = 768;
	}
	else {
	    width = modes[0]->w;
	    height = modes[0]->h;
	}

	if (full_screen) {
	    surface = SDL_SetVideoMode(width, height, 16,
		    SDL_OPENGL|SDL_RESIZABLE|SDL_FULLSCREEN);
	}
	else
	    surface = SDL_SetVideoMode(800, 600, 16, SDL_OPENGL|SDL_RESIZABLE);
	if (!surface)
	    return -1;
	SDL_WM_SetCaption("Fireflies", "fireflies");

	return 0;
    }
#endif
    if (sdl_window)
	cerr << "ERROR: window creation is not supported (compile with HAVE_SDL)" << endl;
    else
	cerr << "ERROR: drawing on root/existing window is not supported (compile with HAVE_GLX)" << endl;
    return -1;
}

int GLCanvas::init()
{
#ifdef WIN32
    srand(time(0));
#else
    struct timeval tv;
    gettimeofday(&tv, 0);
    srand(1000*tv.tv_sec + tv.tv_usec/1000);
#endif

    int ret;
    if ((ret = create_window()) < 0)
	return ret;

    resize();

    return 0;
}

void GLCanvas::resize()
{
#ifdef HAVE_GLX
    if (!sdl_window) {
	XWindowAttributes attrib;
	XGetWindowAttributes(display, window, &attrib);
	width = attrib.width;
	height = attrib.height;
    }
#endif
#ifdef HAVE_SDL
    if (sdl_window) {
	width = surface->w;
	height = surface->h;
    }
#endif
    glViewport(0, 0, width, height);

    scene.resize(width, height);
}

void GLCanvas::draw()
{
    glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

    scene.apply_camera();
    scene.draw();

#ifdef HAVE_GLX
    if (!sdl_window)
	glXSwapBuffers(display, window);
#endif
#ifdef HAVE_SDL
    if (sdl_window)
	SDL_GL_SwapBuffers();
#endif
}

int GLCanvas::loop()
{
    GLuint then, now, millis;
    double seconds;
    bool animate=true;

    then = now = get_millis();

#ifdef DEBUG
    int frames = 0;
    GLuint then_fps = then;
#endif

    while (1) {
	// Draw the scene
	if (need_refresh) {
	    draw();
	    need_refresh = false;
#ifdef DEBUG
	    frames++;
#endif
	}

	now = get_millis();
	millis = now - then;
	seconds = double(millis)/1000.0;

	// Animate the scene
	if (seconds >= spf) {
	    then = now;
	    if (animate) {
		scene.elapse(seconds);
		need_refresh = true;
#ifdef DEBUG
		if (now-then_fps >= 5000) {
		    cerr << frames << " frames in " << now-then_fps << "ms = "
			<< frames/(double(now-then_fps)/1000.0) << endl;
		    then_fps = now;
		    frames = 0;
		}
#endif
	    }
	}
	else {
	    int remain = (int)(1000*(spf - seconds));
	    if (remain > MIN_DELAY) {
#ifdef HAVE_SDL
		SDL_Delay(remain);
#else
		struct timeval tv;
		tv.tv_sec = remain/1000;
		tv.tv_usec = (remain%1000)*1000;
		select(0, 0, 0, 0, &tv);
#endif
	    }
	}

	// Event processing
#ifdef HAVE_GLX
	if (!sdl_window) {
	    XEvent event;
	    while (XPending(display)) {
		XNextEvent(display, &event);
		switch (event.type) {
		case ConfigureNotify:
		    resize();
		    break;
		}
	    }
	}
#endif
#ifdef HAVE_SDL
	if (sdl_window) {
	    SDL_Event event;
	    while (SDL_PollEvent(&event)) {
		switch (event.type) {
		case SDL_MOUSEMOTION:
		    mouse_drag(event.motion);
		    break;
		case SDL_VIDEOEXPOSE:
		    need_refresh = true;
		    break;
		case SDL_VIDEORESIZE:
		    need_refresh = true;
		    surface = SDL_SetVideoMode(event.resize.w, event.resize.h, 16, SDL_OPENGL|SDL_RESIZABLE);
		    if (!surface) {
			cerr << "Weird error: can't set video mode after resize" << endl;
			SDL_Quit();
			return 1;
		    }
		    resize();
		    break;
		case SDL_KEYDOWN:
		    switch (event.key.keysym.sym) {
		    case SDLK_q:
			SDL_Quit();
			return 0;
			break;
		    case SDLK_p: // pause or unpause
			animate = !animate;
			break;
		    case SDLK_t: // show the time
			cout << "Elapsed time: " << scene.curtime << "s" << endl;
			break;
		    case SDLK_UP:
			scene.fast_forward *= 2;
			cout << "fast forward: " << scene.fast_forward << "x" << endl;
			break;
		    case SDLK_DOWN:
			if (scene.fast_forward > 1)
			    scene.fast_forward /= 2;
			cout << "fast forward: " << scene.fast_forward << "x" << endl;
			break;
		    case SDLK_RIGHT:
			scene.fast_forward += 1;
			cout << "fast forward: " << scene.fast_forward << "x" << endl;
			break;
		    case SDLK_LEFT:
			if (scene.fast_forward > 1)
			    scene.fast_forward -= 1;
			cout << "fast forward: " << scene.fast_forward << "x" << endl;
			break;
		    default: {
			int c = event.key.keysym.sym - '0';
			if (event.key.keysym.mod & KMOD_SHIFT) {
			    if (c >= 0 && c < NUM_BMODES) {
				vector<Bait*>::iterator it = scene.baits.begin();
				for (; it != scene.baits.end(); it++)
				    bait_start_mode(*it, c);
				cout << "started bait mode " << c << endl;
			    }
			}
			else if (event.key.keysym.mod & KMOD_CTRL) {
			    if (c >= 0 && c < NUM_BMODES) {
				vector<Bait*>::iterator it = scene.baits.begin();
				for (; it != scene.baits.end(); it++)
				    bait_stop_mode(*it, c);
				cout << "stopped bait mode " << c << endl;
			    }
			}
			else {
			    if (c >= 0 && c < NUM_SMODES) {
				scene_start_mode(c);
				cout << "started scene mode " << c << endl;
			    }
			}
			break;
			     }
		    }
		    break;
		case SDL_QUIT:
		    SDL_Quit();
		    return 0;
		    break;
		}
	    }
	}
#endif
    }
}
#ifdef HAVE_SDL
void GLCanvas::mouse_drag(SDL_MouseMotionEvent& event)
{
    Mat4 inv_camera = rotation_matrix_deg(-scene.camera.rot_angle, scene.camera.rot_axis);
    Vec3 horiz = inv_camera*Vec3(1, 0, 0);
    Vec3 vert = inv_camera*Vec3(0, 1, 0);

    if (SDL_GetModState() & KMOD_CTRL) {
	Quat q = axis_to_quat(scene.camera.rot_axis, DEG_TO_RAD(scene.camera.rot_angle));
	if (event.state & SDL_BUTTON(1)) { // rotate
	    Quat qx = axis_to_quat(vert, event.xrel*0.05);
	    Quat qy = axis_to_quat(horiz, event.yrel*0.05);

	    q = q * qy;
	    q = q * qx;
	    unitize(q);

	    scene.camera.rot_axis = q.vector();
	    scene.camera.rot_angle = RAD_TO_DEG(2.0*acos(q.scalar()));
	    need_refresh = true;
	}
    }
    else {
	if (event.state & SDL_BUTTON(1)) { // move camera
	    scene.camera.pos += 0.5*Vec3(event.xrel, -event.yrel, 0.);
	    need_refresh = true;
	}
	else if (event.state & SDL_BUTTON(3)) { // zoom in/out
	    scene.camera.pos += 0.5*Vec3(0., 0., event.yrel);
	    need_refresh = true;
	}
    }
}
#endif
